/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  SpringProx API
  --------------

  Copyright (c) 2000-2008 SpringCard - www.springcard.com

  sprox_md5.c
  -----------
  Provides the MD5 hash.
 
  Derived from the RSA Data Security, Inc. MD5 Message Digest Algorithm

  revision :
  ----------

  JDA 20/02/2006 : added to the SpringProx library

*/

/**** SpringProxAPI/MD5
 *
 * NAME
 *   SpringProxAPI :: MD5
 *
 * DESCRIPTION
 *   Software implementation of the MD5 hash
 *
 * COPYRIGHT
 *   Derived from the RSA Data Security, Inc. MD5 Message Digest Algorithm
 *
 **/

#include "sprox_api_i.h"

#ifndef SPROX_API_NO_CRYPT

typedef struct
{
  DWORD state[4];
  DWORD count[2];
  BYTE  buffer[64];
} MD5_CTX_ST;

void MD5_Init(MD5_CTX_ST *md5_ctx);
void MD5_Update(MD5_CTX_ST *md5_ctx, const BYTE *input, DWORD size);
void MD5_Final(MD5_CTX_ST *md5_ctx, BYTE digest[16]);

static SPROX_MD5_CTX_ST local_context;

SPRINGPROX_LIB SWORD SPRINGPROX_API SPROX_MD5_Init(SPROX_MD5_CTX_ST *context)
{
  if (context == NULL)
    context = &local_context;
    
  MD5_Init((MD5_CTX_ST *) context);
  return MI_OK;
}

SPRINGPROX_LIB SWORD SPRINGPROX_API SPROX_MD5_Update(SPROX_MD5_CTX_ST *context, const BYTE *inbuf, DWORD size)
{
  if (context == NULL)
    context = &local_context;
   
  if ((inbuf == NULL) || (size == 0))
    return MI_WRONG_PARAMETER_VALUE;
    
  MD5_Update((MD5_CTX_ST *) context, inbuf, size);
  return MI_OK;  
}

SPRINGPROX_LIB SWORD SPRINGPROX_API SPROX_MD5_Final(SPROX_MD5_CTX_ST *context, BYTE *outbuf)
{
  if (context == NULL)
    context = &local_context;
   
  if (outbuf == NULL)
    return MI_WRONG_PARAMETER_VALUE;
    
  MD5_Final((MD5_CTX_ST *) context, outbuf);
  return MI_OK;    
}


/*
 * This is derived work from
 *   "the RSA Data Security, Inc. MD5 Message Digest Algorithm".
 *
 * RSA Data Security, Inc. makes no representations concerning either
 * the merchantability of this software or the suitability of this
 * software for any particular purpose.  It is provided "as is" without
 * express or implied warranty of any kind.
 */

/* MD5 internal functions */
/* ---------------------- */
static void MD5_Transform (DWORD state[4], const BYTE block[64]);
static void MD5_Encode(BYTE *output, DWORD *input, DWORD len);
static void MD5_Decode(DWORD *output, const BYTE *input, DWORD len);

/* MD5 initialization */
/* ------------------ */
void MD5_Init(MD5_CTX_ST *ctx)
{
  /* Current size is 0 */
  ctx->count[0] = ctx->count[1] = 0;
  /* Initial values for state */
  ctx->state[0] = 0x67452301;
  ctx->state[1] = 0xefcdab89;
  ctx->state[2] = 0x98badcfe;
  ctx->state[3] = 0x10325476;
}

/* MD5 block update operation */
/* -------------------------- */
void MD5_Update(MD5_CTX_ST *ctx, const BYTE *input, DWORD inputLen)
{
  DWORD i, index, partLen;

  /* Compute number of bytes mod 64 */
  index = (DWORD)((ctx->count[0] >> 3) & 0x3F);

  /* Update number of bits */
  if ((ctx->count[0] += ((DWORD)inputLen << 3))
      < ((DWORD)inputLen << 3))
    ctx->count[1]++;
  ctx->count[1] += ((DWORD)inputLen >> 29);
  
  partLen = 64 - index;
  
  /* Transform as many times as possible */
  if (inputLen >= partLen)
  {
    memcpy(&ctx->buffer[index], input, partLen);
    MD5_Transform (ctx->state, ctx->buffer);
  
    for (i = partLen; i + 63 < inputLen; i += 64)
      MD5_Transform (ctx->state, &input[i]);
    
    index = 0;

  } else
    i = 0;
  
  /* Store remaining input in buffer */
  memcpy(&ctx->buffer[index], &input[i], inputLen-i);
}

/* MD5 finalization */
/* ---------------- */
void MD5_Final(MD5_CTX_ST *ctx, BYTE digest[16])
{
  BYTE bits[8];
  DWORD index, padLen;
  static BYTE PADDING[64] = {
    0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
  };

  /* Save number of bits */
  MD5_Encode(bits, ctx->count, 8);

  /* Pad out to 56 mod 64 */
  index = (DWORD)((ctx->count[0] >> 3) & 0x3f);
  padLen = (index < 56) ? (56 - index) : (120 - index);
  MD5_Update(ctx, PADDING, padLen);
  
  /* Append length (before padding) */
  MD5_Update(ctx, bits, 8);

  /* Store state in digest */
  MD5_Encode(digest, ctx->state, 16);
}


/*
 * The MD5 transformation core
 * ---------------------------
 */

/* Constants for the MD5 transformation */
#define S11 7
#define S12 12
#define S13 17
#define S14 22
#define S21 5
#define S22 9
#define S23 14
#define S24 20
#define S31 4
#define S32 11
#define S33 16
#define S34 23
#define S41 6
#define S42 10
#define S43 15
#define S44 21

/* ROTATE_LEFT rotates x left n bits */
#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* Basic functions F, G, H and I */
#define MD5_F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define MD5_G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define MD5_H(x, y, z) ((x) ^ (y) ^ (z))
#define MD5_I(x, y, z) ((y) ^ ((x) | (~z)))

/* MD5_FF, MD5_GG, MD5_HH, and MD5_II transformations for rounds 1, 2, 3, and 4
   Rotation is separate from addition to prevent recomputation */
#define MD5_FF(a, b, c, d, x, s, ac) {(a) += MD5_F ((b), (c), (d)) + (x) + (DWORD)(ac); (a) = ROTATE_LEFT ((a), (s)); (a) += (b); }
#define MD5_GG(a, b, c, d, x, s, ac) {(a) += MD5_G ((b), (c), (d)) + (x) + (DWORD)(ac); (a) = ROTATE_LEFT ((a), (s)); (a) += (b); }
#define MD5_HH(a, b, c, d, x, s, ac) {(a) += MD5_H ((b), (c), (d)) + (x) + (DWORD)(ac); (a) = ROTATE_LEFT ((a), (s)); (a) += (b); }
#define MD5_II(a, b, c, d, x, s, ac) {(a) += MD5_I ((b), (c), (d)) + (x) + (DWORD)(ac); (a) = ROTATE_LEFT ((a), (s)); (a) += (b); }

/* MD5 basic transformation. Transforms state based on block */
static void MD5_Transform (DWORD state[4], const BYTE block[64])
{
  DWORD a = state[0], b = state[1], c = state[2], d = state[3], x[16];
  
  MD5_Decode(x, block, 64);

  /* Round 1 */
  MD5_FF (a, b, c, d, x[ 0], S11, 0xd76aa478); /* 1 */
  MD5_FF (d, a, b, c, x[ 1], S12, 0xe8c7b756); /* 2 */
  MD5_FF (c, d, a, b, x[ 2], S13, 0x242070db); /* 3 */
  MD5_FF (b, c, d, a, x[ 3], S14, 0xc1bdceee); /* 4 */
  MD5_FF (a, b, c, d, x[ 4], S11, 0xf57c0faf); /* 5 */
  MD5_FF (d, a, b, c, x[ 5], S12, 0x4787c62a); /* 6 */
  MD5_FF (c, d, a, b, x[ 6], S13, 0xa8304613); /* 7 */
  MD5_FF (b, c, d, a, x[ 7], S14, 0xfd469501); /* 8 */
  MD5_FF (a, b, c, d, x[ 8], S11, 0x698098d8); /* 9 */
  MD5_FF (d, a, b, c, x[ 9], S12, 0x8b44f7af); /* 10 */
  MD5_FF (c, d, a, b, x[10], S13, 0xffff5bb1); /* 11 */
  MD5_FF (b, c, d, a, x[11], S14, 0x895cd7be); /* 12 */
  MD5_FF (a, b, c, d, x[12], S11, 0x6b901122); /* 13 */
  MD5_FF (d, a, b, c, x[13], S12, 0xfd987193); /* 14 */
  MD5_FF (c, d, a, b, x[14], S13, 0xa679438e); /* 15 */
  MD5_FF (b, c, d, a, x[15], S14, 0x49b40821); /* 16 */

  /* Round 2 */
  MD5_GG (a, b, c, d, x[ 1], S21, 0xf61e2562); /* 17 */
  MD5_GG (d, a, b, c, x[ 6], S22, 0xc040b340); /* 18 */
  MD5_GG (c, d, a, b, x[11], S23, 0x265e5a51); /* 19 */
  MD5_GG (b, c, d, a, x[ 0], S24, 0xe9b6c7aa); /* 20 */
  MD5_GG (a, b, c, d, x[ 5], S21, 0xd62f105d); /* 21 */
  MD5_GG (d, a, b, c, x[10], S22,  0x2441453); /* 22 */
  MD5_GG (c, d, a, b, x[15], S23, 0xd8a1e681); /* 23 */
  MD5_GG (b, c, d, a, x[ 4], S24, 0xe7d3fbc8); /* 24 */
  MD5_GG (a, b, c, d, x[ 9], S21, 0x21e1cde6); /* 25 */
  MD5_GG (d, a, b, c, x[14], S22, 0xc33707d6); /* 26 */
  MD5_GG (c, d, a, b, x[ 3], S23, 0xf4d50d87); /* 27 */
  MD5_GG (b, c, d, a, x[ 8], S24, 0x455a14ed); /* 28 */
  MD5_GG (a, b, c, d, x[13], S21, 0xa9e3e905); /* 29 */
  MD5_GG (d, a, b, c, x[ 2], S22, 0xfcefa3f8); /* 30 */
  MD5_GG (c, d, a, b, x[ 7], S23, 0x676f02d9); /* 31 */
  MD5_GG (b, c, d, a, x[12], S24, 0x8d2a4c8a); /* 32 */

  /* Round 3 */
  MD5_HH (a, b, c, d, x[ 5], S31, 0xfffa3942); /* 33 */
  MD5_HH (d, a, b, c, x[ 8], S32, 0x8771f681); /* 34 */
  MD5_HH (c, d, a, b, x[11], S33, 0x6d9d6122); /* 35 */
  MD5_HH (b, c, d, a, x[14], S34, 0xfde5380c); /* 36 */
  MD5_HH (a, b, c, d, x[ 1], S31, 0xa4beea44); /* 37 */
  MD5_HH (d, a, b, c, x[ 4], S32, 0x4bdecfa9); /* 38 */
  MD5_HH (c, d, a, b, x[ 7], S33, 0xf6bb4b60); /* 39 */
  MD5_HH (b, c, d, a, x[10], S34, 0xbebfbc70); /* 40 */
  MD5_HH (a, b, c, d, x[13], S31, 0x289b7ec6); /* 41 */
  MD5_HH (d, a, b, c, x[ 0], S32, 0xeaa127fa); /* 42 */
  MD5_HH (c, d, a, b, x[ 3], S33, 0xd4ef3085); /* 43 */
  MD5_HH (b, c, d, a, x[ 6], S34,  0x4881d05); /* 44 */
  MD5_HH (a, b, c, d, x[ 9], S31, 0xd9d4d039); /* 45 */
  MD5_HH (d, a, b, c, x[12], S32, 0xe6db99e5); /* 46 */
  MD5_HH (c, d, a, b, x[15], S33, 0x1fa27cf8); /* 47 */
  MD5_HH (b, c, d, a, x[ 2], S34, 0xc4ac5665); /* 48 */

  /* Round 4 */
  MD5_II (a, b, c, d, x[ 0], S41, 0xf4292244); /* 49 */
  MD5_II (d, a, b, c, x[ 7], S42, 0x432aff97); /* 50 */
  MD5_II (c, d, a, b, x[14], S43, 0xab9423a7); /* 51 */
  MD5_II (b, c, d, a, x[ 5], S44, 0xfc93a039); /* 52 */
  MD5_II (a, b, c, d, x[12], S41, 0x655b59c3); /* 53 */
  MD5_II (d, a, b, c, x[ 3], S42, 0x8f0ccc92); /* 54 */
  MD5_II (c, d, a, b, x[10], S43, 0xffeff47d); /* 55 */
  MD5_II (b, c, d, a, x[ 1], S44, 0x85845dd1); /* 56 */
  MD5_II (a, b, c, d, x[ 8], S41, 0x6fa87e4f); /* 57 */
  MD5_II (d, a, b, c, x[15], S42, 0xfe2ce6e0); /* 58 */
  MD5_II (c, d, a, b, x[ 6], S43, 0xa3014314); /* 59 */
  MD5_II (b, c, d, a, x[13], S44, 0x4e0811a1); /* 60 */
  MD5_II (a, b, c, d, x[ 4], S41, 0xf7537e82); /* 61 */
  MD5_II (d, a, b, c, x[11], S42, 0xbd3af235); /* 62 */
  MD5_II (c, d, a, b, x[ 2], S43, 0x2ad7d2bb); /* 63 */
  MD5_II (b, c, d, a, x[ 9], S44, 0xeb86d391); /* 64 */

  state[0] += a;
  state[1] += b;
  state[2] += c;
  state[3] += d;
  
  /* Zeroize sensitive information */
  memset (x, 0, sizeof (x));
}

/* Encodes input (DWORD) into output (BYTE). Assumes len is a multiple of 4 */
static void MD5_Encode (BYTE *output, DWORD *input, DWORD len)
{
  DWORD i, j;

  for (i = 0, j = 0; j < len; i++, j += 4) {
    output[j] = (BYTE)(input[i] & 0xff);
    output[j+1] = (BYTE)((input[i] >> 8) & 0xff);
    output[j+2] = (BYTE)((input[i] >> 16) & 0xff);
    output[j+3] = (BYTE)((input[i] >> 24) & 0xff);
  }
}

/* Decodes input (BYTE) into output (DWORD). Assumes len is a multiple of 4 */
static void MD5_Decode (DWORD *output, const BYTE *input, DWORD len)
{
  DWORD i, j;

  for (i = 0, j = 0; j < len; i++, j += 4)
    output[i] = ((DWORD)input[j]) | (((DWORD)input[j+1]) << 8) |
      (((DWORD)input[j+2]) << 16) | (((DWORD)input[j+3]) << 24);
}

#endif
